/*
Activity Indicators for Modern C++
https://github.com/p-ranav/indicators

Licensed under the MIT License <http://opensource.org/licenses/MIT>.
SPDX-License-Identifier: MIT
Copyright (c) 2019 Pranav Srinivas Kumar <pranav.srinivas.kumar@gmail.com>.

Permission is hereby  granted, free of charge, to any  person obtaining a copy
of this software and associated  documentation files (the "Software"), to deal
in the Software  without restriction, including without  limitation the rights
to  use, copy,  modify, merge,  publish, distribute,  sublicense, and/or  sell
copies  of  the Software,  and  to  permit persons  to  whom  the Software  is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE  IS PROVIDED "AS  IS", WITHOUT WARRANTY  OF ANY KIND,  EXPRESS OR
IMPLIED,  INCLUDING BUT  NOT  LIMITED TO  THE  WARRANTIES OF  MERCHANTABILITY,
FITNESS FOR  A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT  SHALL THE
AUTHORS  OR COPYRIGHT  HOLDERS  BE  LIABLE FOR  ANY  CLAIM,  DAMAGES OR  OTHER
LIABILITY, WHETHER IN AN ACTION OF  CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE  OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/
#pragma once

#include <indicators/color.hpp>
#include <indicators/details/stream_helper.hpp>

#define NOMINMAX
#include <algorithm>
#include <atomic>
#include <chrono>
#include <iomanip>
#include <iostream>
#include <mutex>
#include <string>
#include <tuple>
#include <thread>
#include "setting.hpp"

namespace indicators {

class BlockProgressBar {
  using Settings = std::tuple<
      option::ForegroundColor,
      option::BarWidth,
      option::Start,
      option::End,
      option::PrefixText,
      option::PostfixText,
      option::ShowPercentage,
      option::ShowElapsedTime,
      option::ShowRemainingTime,
      option::Completed,
      option::SavedStartTime,
      option::MaxPostfixTextLen>;
public:
  template <typename... Args, typename std::enable_if<details::are_settings_from_tuple<Settings, typename std::decay<Args>::type...>::value, void*>::type = nullptr>
  explicit BlockProgressBar(Args&&... args) : settings_(
      details::get<details::ProgressBarOption::foreground_color>(option::ForegroundColor{Color::WHITE}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::bar_width>(option::BarWidth{100}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::start>(option::Start{"["}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::end>(option::End{"]"}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::prefix_text>(option::PrefixText{""}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::postfix_text>(option::PostfixText{""}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::show_percentage>(option::ShowPercentage{true}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::show_elapsed_time>(option::ShowElapsedTime{false}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::show_remaining_time>(option::ShowRemainingTime{false}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::completed>(option::Completed{false}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::saved_start_time>(option::SavedStartTime{false}, std::forward<Args>(args)...),
      details::get<details::ProgressBarOption::max_postfix_text_len>(option::MaxPostfixTextLen{0}, std::forward<Args>(args)...)
      ) {}

  template <typename T, details::ProgressBarOption id>
  void set_option(details::Setting<T, id>&& setting){
    static_assert(!std::is_same<T, typename std::decay<decltype(details::get_value<id>(std::declval<Settings>()))>::type>::value, "Setting has wrong type!");
    std::lock_guard<std::mutex> lock(_mutex);
    get_value<id>() = std::move(setting).value;
  }

  template <typename T, details::ProgressBarOption id>
  void set_option(const details::Setting<T, id>& setting){
    static_assert(!std::is_same<T, typename std::decay<decltype(details::get_value<id>(std::declval<Settings>()))>::type>::value, "Setting has wrong type!");
    std::lock_guard<std::mutex> lock(_mutex);
    get_value<id>() = setting.value;
  }

  void set_option(const details::Setting<std::string, details::ProgressBarOption::postfix_text>& setting){
    std::lock_guard<std::mutex> lock(_mutex);
    get_value<details::ProgressBarOption::postfix_text>() = setting.value;
    if(setting.value.length() > get_value<details::ProgressBarOption::max_postfix_text_len>()){
      get_value<details::ProgressBarOption::max_postfix_text_len>() = setting.value.length();
    }
  }

  void set_option(details::Setting<std::string, details::ProgressBarOption::postfix_text>&& setting){
    std::lock_guard<std::mutex> lock(_mutex);
    get_value<details::ProgressBarOption::postfix_text>() = std::move(setting).value;
    auto& new_value = get_value<details::ProgressBarOption::postfix_text>();
    if(new_value.length() > get_value<details::ProgressBarOption::max_postfix_text_len>()){
      get_value<details::ProgressBarOption::max_postfix_text_len>() = new_value.length();
    }
  }

  void set_progress(float value) {
    {
      std::lock_guard<std::mutex> lock{_mutex};
      _progress = value;
    }
    _save_start_time();
    _print_progress();
  }

  void tick() {
    {
      std::lock_guard<std::mutex> lock{_mutex};
      _progress += 1;
    }
    _save_start_time();
    _print_progress();
  }

  size_t current() {
    std::lock_guard<std::mutex> lock{_mutex};
    return std::min(static_cast<size_t>(_progress), size_t(100));
  }

  bool is_completed() const { return get_value<details::ProgressBarOption::completed>(); }

  void mark_as_completed() {
    get_value<details::ProgressBarOption::completed>() = true;
    _print_progress();
  }

private:

  template <details::ProgressBarOption id>
  auto get_value() -> decltype((details::get_value<id>(std::declval<Settings&>()).value)) {
    return details::get_value<id>(settings_).value;
  }

  template <details::ProgressBarOption id>
  auto get_value() const -> decltype((details::get_value<id>(std::declval<const Settings&>()).value)) {
    return details::get_value<id>(settings_).value;
  }

  Settings settings_;
  float _progress{0.0};
  std::chrono::time_point<std::chrono::high_resolution_clock> _start_time_point;
  std::mutex _mutex;

  template <typename Indicator, size_t count> friend class MultiProgress;
  std::atomic<bool> _multi_progress_mode{false};

  void _save_start_time() {
    auto& show_elapsed_time = get_value<details::ProgressBarOption::show_elapsed_time>();
    auto& saved_start_time = get_value<details::ProgressBarOption::saved_start_time>();
    auto& show_remaining_time = get_value<details::ProgressBarOption::show_remaining_time>();
    if ((show_elapsed_time || show_remaining_time) && !saved_start_time) {
      _start_time_point = std::chrono::high_resolution_clock::now();
      saved_start_time = true;
    }
  }

  void _print_progress(bool from_multi_progress = false) {
    if (_multi_progress_mode && !from_multi_progress) {
      if (_progress > 100.0) {
        get_value<details::ProgressBarOption::completed>() = true;
      }
      return;
    }
    std::lock_guard<std::mutex> lock{_mutex};
    auto now = std::chrono::high_resolution_clock::now();
    auto elapsed = std::chrono::duration_cast<std::chrono::nanoseconds>(now - _start_time_point);

    std::cout << termcolor::bold;
    details::set_stream_color(std::cout, get_value<details::ProgressBarOption::foreground_color>());
    std::cout << get_value<details::ProgressBarOption::prefix_text>();
    std::cout << get_value<details::ProgressBarOption::start>();

    details::BlockProgressScaleWriter writer{std::cout, get_value<details::ProgressBarOption::bar_width>()};
    writer.write(_progress);

    std::cout << get_value<details::ProgressBarOption::end>();
    if (get_value<details::ProgressBarOption::show_percentage>()) {
      std::cout << " " << std::min(static_cast<size_t>(_progress), size_t(100)) << "%";
    }

    if (get_value<details::ProgressBarOption::show_elapsed_time>()) {
      std::cout << " [";
      details::write_duration(std::cout, elapsed);
    }

    if (get_value<details::ProgressBarOption::show_remaining_time>()) {
      if (get_value<details::ProgressBarOption::show_elapsed_time>())
        std::cout << "<";
      else
        std::cout << " [";
      auto eta = std::chrono::nanoseconds(
          _progress > 0 ? static_cast<long long>(elapsed.count() * 100 / _progress) : 0);
      auto remaining = eta > elapsed ? (eta - elapsed) : (elapsed - eta);
      details::write_duration(std::cout, remaining);
      std::cout << "]";
    } else {
      if (get_value<details::ProgressBarOption::show_elapsed_time>())
        std::cout << "]";
    }

    if (get_value<details::ProgressBarOption::max_postfix_text_len>() == 0)
      get_value<details::ProgressBarOption::max_postfix_text_len>() = 10;
    std::cout << " " << get_value<details::ProgressBarOption::postfix_text>() << std::string(get_value<details::ProgressBarOption::max_postfix_text_len>(), ' ') << "\r";
    std::cout.flush();
    if (_progress > 100.0) {
      get_value<details::ProgressBarOption::completed>() = true;
    }
    if (get_value<details::ProgressBarOption::completed>() && !from_multi_progress) // Don't std::endl if calling from MultiProgress
      std::cout << termcolor::reset << std::endl;
  }
};

} // namespace indicators
